@@ -36,19 +36,17 @@
 #include <sys/types.h>
 #include <sys/stat.h>
 #include <sys/statvfs.h>
+#include <sys/sendfile.h>
 #include <fcntl.h>
 #include <wait.h>
+#include <inttypes.h>
 
 #include <glib.h>
 
-#include <openobex/obex.h>
-#include <openobex/obex_const.h>
-
+#include "obexd.h"
 #include "plugin.h"
 #include "log.h"
-#include "obex.h"
 #include "mimetype.h"
-#include "service.h"
 #include "filesystem.h"
 
 #define EOL_CHARS "\n"
@@ -77,15 +75,50 @@
 			" modified=\"%s\" mem-type=\"DEV\"" \
 			" created=\"%s\"/>" EOL_CHARS
 
-static const uint8_t FTP_TARGET[TARGET_SIZE] = {
+#define FTP_TARGET_SIZE 16
+
+static const uint8_t FTP_TARGET[FTP_TARGET_SIZE] = {
 			0xF9, 0xEC, 0x7B, 0xC4,  0x95, 0x3C, 0x11, 0xD2,
 			0x98, 0x4E, 0x52, 0x54,  0x00, 0xDC, 0x9E, 0x09  };
 
 #define PCSUITE_WHO_SIZE 8
 
 static const uint8_t PCSUITE_WHO[PCSUITE_WHO_SIZE] = {
-			'P','C',' ','S','u','i','t','e' };
+			'P', 'C', ' ', 'S', 'u', 'i', 't', 'e' };
 
+gboolean is_filename(const char *name)
+{
+	if (strchr(name, '/'))
+		return FALSE;
+
+	if (strcmp(name, ".") == 0)
+		return FALSE;
+
+	if (strcmp(name, "..") == 0)
+		return FALSE;
+
+	return TRUE;
+}
+
+int verify_path(const char *path)
+{
+	char *t;
+	int ret = 0;
+
+	if (obex_option_symlinks())
+		return 0;
+
+	t = realpath(path, NULL);
+	if (t == NULL)
+		return -errno;
+
+	if (!g_str_has_prefix(t, obex_option_root_folder()))
+		ret = -EPERM;
+
+	free(t);
+
+	return ret;
+}
 
 static char *file_stat_line(char *filename, struct stat *fstat,
 					struct stat *dstat, gboolean root,
@@ -135,12 +168,10 @@ static void *filesystem_open(const char *name, int oflag, mode_t mode,
 {
 	struct stat stats;
 	struct statvfs buf;
-	const char *root_folder;
-	char *folder;
-	gboolean root;
-	int fd = open(name, oflag, mode);
+	int fd, ret;
 	uint64_t avail;
 
+	fd = open(name, oflag, mode);
 	if (fd < 0) {
 		if (err)
 			*err = -errno;
@@ -153,19 +184,11 @@ static void *filesystem_open(const char *name, int oflag, mode_t mode,
 		goto failed;
 	}
 
-	root_folder = obex_option_root_folder();
-	folder = g_path_get_dirname(name);
-	root = g_strcmp0(folder, root_folder);
-
-	g_free(folder);
-
-	if (!root || obex_option_symlinks()) {
-		if (S_ISLNK(stats.st_mode)) {
-			if (err)
-				*err = -EPERM;
-			goto failed;
-		}
-
+	ret = verify_path(name);
+	if (ret < 0) {
+		if (err)
+			*err = ret;
+		goto failed;
 	}
 
 	if (oflag == O_RDONLY) {
@@ -209,8 +232,7 @@ static int filesystem_close(void *object)
 	return 0;
 }
 
-static ssize_t filesystem_read(void *object, void *buf, size_t count,
-								uint8_t *hi)
+static ssize_t filesystem_read(void *object, void *buf, size_t count)
 {
 	ssize_t ret;
 
@@ -218,8 +240,6 @@ static ssize_t filesystem_read(void *object, void *buf, size_t count,
 	if (ret < 0)
 		return -errno;
 
-	*hi = OBEX_HDR_BODY;
-
 	return ret;
 }
 
@@ -234,6 +254,99 @@ static ssize_t filesystem_write(void *object, const void *buf, size_t count)
 	return ret;
 }
 
+static int filesystem_rename(const char *name, const char *destname)
+{
+	int ret;
+
+	ret = rename(name, destname);
+	if (ret < 0) {
+		error("rename(%s, %s): %s (%d)", name, destname,
+						strerror(errno), errno);
+		return -errno;
+	}
+
+	return ret;
+}
+
+static int sendfile_async(int out_fd, int in_fd, off_t *offset, size_t count)
+{
+	int pid;
+
+	/* Run sendfile on child process */
+	pid = fork();
+	switch (pid) {
+		case 0:
+			break;
+		case -1:
+			error("fork() %s (%d)", strerror(errno), errno);
+			return -errno;
+		default:
+			DBG("child %d forked", pid);
+			return pid;
+	}
+
+	/* At child */
+	if (sendfile(out_fd, in_fd, offset, count) < 0)
+		error("sendfile(): %s (%d)", strerror(errno), errno);
+
+	close(in_fd);
+	close(out_fd);
+
+	exit(errno);
+}
+
+static int filesystem_copy(const char *name, const char *destname)
+{
+	void *in, *out;
+	ssize_t ret;
+	size_t size;
+	struct stat st;
+	int in_fd, out_fd, err;
+
+	in = filesystem_open(name, O_RDONLY, 0, NULL, &size, &err);
+	if (in == NULL) {
+		error("open(%s): %s (%d)", name, strerror(-err), -err);
+		return -err;
+	}
+
+	in_fd = GPOINTER_TO_INT(in);
+	ret = fstat(in_fd, &st);
+	if (ret < 0) {
+		error("stat(%s): %s (%d)", name, strerror(errno), errno);
+		return -errno;
+	}
+
+	out = filesystem_open(destname, O_WRONLY | O_CREAT | O_TRUNC,
+					st.st_mode, NULL, &size, &err);
+	if (out == NULL) {
+		error("open(%s): %s (%d)", destname, strerror(-err), -err);
+		filesystem_close(in);
+		return -errno;
+	}
+
+	out_fd = GPOINTER_TO_INT(out);
+
+	/* Check if sendfile is supported */
+	ret = sendfile(out_fd, in_fd, NULL, 0);
+	if (ret < 0) {
+		ret = -errno;
+		error("sendfile: %s (%zd)", strerror(-ret), -ret);
+		goto done;
+	}
+
+	ret = sendfile_async(out_fd, in_fd, NULL, st.st_size);
+	if (ret < 0)
+		goto done;
+
+	return 0;
+
+done:
+	filesystem_close(in);
+	filesystem_close(out);
+
+	return ret;
+}
+
 struct capability_object {
 	int pid;
 	int output;
@@ -363,7 +476,7 @@ static GString *append_listing(GString *object, const char *name,
 	struct stat fstat, dstat;
 	struct dirent *ep;
 	DIR *dp;
-	gboolean root, symlinks;
+	gboolean root;
 	int ret;
 
 	root = g_str_equal(name, obex_option_root_folder());
@@ -375,14 +488,16 @@ static GString *append_listing(GString *object, const char *name,
 		goto failed;
 	}
 
-	symlinks = obex_option_symlinks();
-	if (root && symlinks)
-		ret = stat(name, &dstat);
-	else {
+	if (!root)
 		object = g_string_append(object, FL_PARENT_FOLDER_ELEMENT);
-		ret = lstat(name, &dstat);
+
+	ret = verify_path(name);
+	if (ret < 0) {
+		*err = ret;
+		goto failed;
 	}
 
+	ret = stat(name, &dstat);
 	if (ret < 0) {
 		if (err)
 			*err = -errno;
@@ -398,21 +513,16 @@ static GString *append_listing(GString *object, const char *name,
 			continue;
 
 		filename = g_filename_to_utf8(ep->d_name, -1, NULL, NULL, NULL);
-		if (name == NULL) {
+		if (filename == NULL) {
 			error("g_filename_to_utf8: invalid filename");
 			continue;
 		}
 
 		fullname = g_build_filename(name, ep->d_name, NULL);
 
-		if (root && symlinks)
-			ret = stat(fullname, &fstat);
-		else
-			ret = lstat(fullname, &fstat);
-
+		ret = stat(fullname, &fstat);
 		if (ret < 0) {
-			DBG("%s: %s(%d)", root ? "stat" : "lstat",
-					strerror(errno), errno);
+			DBG("stat: %s(%d)", strerror(errno), errno);
 			g_free(filename);
 			g_free(fullname);
 			continue;
@@ -494,24 +604,20 @@ ssize_t string_read(void *object, void *buf, size_t count)
 
 	len = MIN(string->len, count);
 	memcpy(buf, string->str, len);
-	string = g_string_erase(string, 0, len);
+	g_string_erase(string, 0, len);
 
 	return len;
 }
 
-static ssize_t folder_read(void *object, void *buf, size_t count, uint8_t *hi)
+static ssize_t folder_read(void *object, void *buf, size_t count)
 {
-	*hi = OBEX_HDR_BODY;
 	return string_read(object, buf, count);
 }
 
-static ssize_t capability_read(void *object, void *buf, size_t count,
-								uint8_t *hi)
+static ssize_t capability_read(void *object, void *buf, size_t count)
 {
 	struct capability_object *obj = object;
 
-	*hi = OBEX_HDR_BODY;
-
 	if (obj->buffer)
 		return string_read(obj->buffer, buf, count);
 
@@ -555,11 +661,13 @@ static struct obex_mime_type_driver file = {
 	.read = filesystem_read,
 	.write = filesystem_write,
 	.remove = remove,
+	.move = filesystem_rename,
+	.copy = filesystem_copy,
 };
 
 static struct obex_mime_type_driver capability = {
 	.target = FTP_TARGET,
-	.target_size = TARGET_SIZE,
+	.target_size = FTP_TARGET_SIZE,
 	.mimetype = "x-obex/capability",
 	.open = capability_open,
 	.close = capability_close,
@@ -568,7 +676,7 @@ static struct obex_mime_type_driver capability = {
 
 static struct obex_mime_type_driver folder = {
 	.target = FTP_TARGET,
-	.target_size = TARGET_SIZE,
+	.target_size = FTP_TARGET_SIZE,
 	.mimetype = "x-obex/folder-listing",
 	.open = folder_open,
 	.close = string_free,
@@ -577,7 +685,7 @@ static struct obex_mime_type_driver folder = {
 
 static struct obex_mime_type_driver pcsuite = {
 	.target = FTP_TARGET,
-	.target_size = TARGET_SIZE,
+	.target_size = FTP_TARGET_SIZE,
 	.who = PCSUITE_WHO,
 	.who_size = PCSUITE_WHO_SIZE,
 	.mimetype = "x-obex/folder-listing",
