@@ -49,7 +49,6 @@
 #include "mimetype.h"
 #include "filesystem.h"
 #include "manager.h"
-#include "glib-helper.h"
 
 #define PHONEBOOK_TYPE		"x-bt/phonebook"
 #define VCARDLISTING_TYPE	"x-bt/vcard-listing"
@@ -962,7 +961,7 @@ static ssize_t array_read(GByteArray *array, void *buf, size_t count)
 
 	len = MIN(array->len, count);
 	memcpy(buf, array->data, len);
-	array = g_byte_array_remove_range(array, 0, len);
+	g_byte_array_remove_range(array, 0, len);
 
 	return len;
 }
